#ifndef BITMAP_H
#define BITMAP_H

#ifndef HOST_NUM_REGS
#error HOST_NUM_REGS not set
#endif

#if HOST_NUM_REGS <= 64

#if HOST_NUM_REGS <= 32
typedef uint32_t host_reg_bitmap;
#else
typedef uint64_t host_reg_bitmap;
#endif

#define HBM_CLEAR(dext) (dext) = 0
#define HBM_AND(dest, src1, src2) (dest) = (src1) & (src2)
#define HBM_OR(dest, src1, src2) (dest) = (src1) | (src2)
#define HBM_BIC(dest, src1, src2) (dest) = (src1) & ~(src2)
#define HBM_COPY(dest, src) (dest) = (src)
#define HBM_IS_EMPTY(src) ((src) == 0)
#define HBM_SET_REG(dest, regno) (dest) |= ((host_reg_bitmap)1 << (regno))
#define HBM_CLEAR_REG(dest, regno) (dest) &= ~((host_reg_bitmap)1 << (regno))
#define HBM_TEST_REG(dest, regno) \
    (((dest) & ((host_reg_bitmap)1 << (regno))) != 0)
#define HBM_IS_SINGLE_REG(src) (((src) & ((src) - 1)) == 0)

#else
/* More than 64 regs.  */
#error too many regs
#define HBM_WORDS (HOST_NUM_REGS / 32)
/* This is declared as an array so that declarations work, and function
   arguments decay to pointers.  */
typedef uint32_t host_reg_bitmap[HBM_WORDS];
#define HBMP_PTR 

static inline void HBM_CLEAR(host_reg_bitmap dest)
{
    memset(dest, 0, sizeof(host_reg_bitmap));
}

static inline void HBM_AND(host_reg_bitmap dest,
                           host_reg_bitmap src1, 
                           host_reg_bitmap src2)
{
    int n;
    for (n = 0; n < HBM_WORDS; n++)
        dest[n] = src1[n] & src2[n];
}

static inline void HBM_OR(host_reg_bitmap dest,
                          host_reg_bitmap src1, 
                          host_reg_bitmap src2)
{
    int n;
    for (n = 0; n < HBM_WORDS; n++)
        dest[n] = src1[n] | src2[n];
}

static inline void HBM_BIC(host_reg_bitmap dest,
                           host_reg_bitmap src1, 
                           host_reg_bitmap src2)
{
    int n;
    for (n = 0; n < HBM_WORDS; n++)
        dest[n] = src1[n] &~ src2[n];
}

static inline void HBM_COPY(host_reg_bitmap dest,
                            host_reg_bitmap src)
{
    int n;
    for (n = 0; n < HBM_WORDS; n++)
        dest[n] = src[n];
}

static inline int HBM_IS_EMPTY(host_reg_bitmap src)
{
    int n;
    for (n = 0; n < HBM_WORDS; n++) {
        if (src[n])
            return 0;
    }
    return 1;
}

static inline void HBM_SET_REG(host_reg_bitmap dest, int regno)
{
    int n;
    n = regno >> 5;
    dest[n] |= ((host_reg_bitmap)1 << (regno & 31));
}

static inline void HBM_CLEAR_REG(host_reg_bitmap dest, int regno)
{
    int n;
    n = regno >> 5;
    dest[n] &= ~((host_reg_bitmap)1 << (regno & 31));
}

static inline int HBM_TEST_REG(host_reg_bitmap dest, int regno)
{
    int n;
    n = regno >> 5;
    if (dest[n] & ((host_reg_bitmap)1 << (regno & 31)))
        return 1;
    else
        return 0;
}

static inline int HBM_IS_SINGLE_REG(host_reg_bitmap src)
{
    int n;
    int seen_bit = 0;
    for (n = 0; n < HBM_WORDS; n++) {
        if (seen_bit && src[n] != 0)
            return 0;
        if (((src[n]) & ((src[n]) - 1)) != 0)
            return 0;
        if (src[0])
            seen_bit = 1;
    }
    /* Should never happen.  */
    return 0;
}

#endif

/* This is a lame implementation. Can probably optimize lots.  */
static inline int HBM_FIRST_REGNO(host_reg_bitmap src)
{
    int n;
    for (n = 0; n < 32; n++) {
        if (HBM_TEST_REG(src, n))
            return n;
    }
    /* Should never happen.  */
    return -1;
}

extern host_reg_bitmap allocatable_regs;
extern host_reg_bitmap host_reg_classes[NUM_HOST_REG_CLASSES];

void init_host_reg_classes(void);

#endif

