static inline int gen_new_label(void)
{
    return nb_gen_labels++;
}

static inline void gen_set_label(int n)
{
    gen_op_label(n);
}

#if HOST_LONG_BITS < 64
/* Routines to lower 64-bit qops on 32-bit hosts.  */
static inline int highpart(int qreg)
{
  return gen_reginfo[qreg].highpart;
}

#define SPLIT_UNARY(name) \
static inline void gen_op_##name##64(long arg0, long arg1) \
{ \
    gen_op_##name##32(arg0, arg1); \
    gen_op_##name##32(highpart(arg0), highpart(arg1)); \
}

SPLIT_UNARY(mov)
SPLIT_UNARY(not)

static inline void gen_op_neg64(long arg0, long arg1)
{
    gen_op_neg64_split(arg0, highpart(arg0), arg1, highpart(arg1));
}

static inline void gen_op_bswap64(long arg0, long arg1)
{
    if (arg0 == arg1) {
        int tmp;
        tmp = gen_new_qreg(QMODE_I32);
        gen_op_bswap32(tmp, highpart(arg1));
        gen_op_bswap32(highpart(arg0), arg1);
        gen_op_mov32(arg0, tmp);
    } else {
        gen_op_bswap32(arg0, highpart(arg1));
        gen_op_bswap32(highpart(arg0), arg1);
    }
}

#define SPLIT_BINARY(name) \
static inline void gen_op_##name##64(long arg0, long arg1, long arg2) \
{ \
    gen_op_##name##32(arg0, arg1, arg2); \
    gen_op_##name##32(highpart(arg0), highpart(arg1), \
                      highpart(arg2)); \
}

SPLIT_BINARY(and)
SPLIT_BINARY(or)
SPLIT_BINARY(xor)

static inline void gen_op_add64(long arg0, long arg1, long arg2)
{
    gen_op_add64_split(arg0, highpart(arg0),
                       arg1, highpart(arg1),
                       arg2, highpart(arg2));
}

static inline void gen_op_sub64(long arg0, long arg1, long arg2)
{
    gen_op_sub64_split(arg0, highpart(arg0),
                       arg1, highpart(arg1),
                       arg2, highpart(arg2));
}

static inline void gen_split_shift32(uint32_t dest, uint32_t src,
                                     uint32_t count, int right, int arith)
{
    if (right) {
        if (arith)
            gen_op_sar32(dest, src, count);
        else
            gen_op_shr32(dest, src, count);
    } else {
        gen_op_shl32(dest, src, count);
    }
}

static inline void split_shift64(long arg0, long arg1, long arg2,
                                 int right, int arithmetic)
{
    int dest_top, dest_bot;
    int src_top, src_bot;
    int tmp;
    int carry;
    int end_label;
    int small_label;
    int nonzero_label;

    /* Shift is from top to bot.  */
    if (right) {
        dest_top = highpart(arg0);
        dest_bot = arg0;
        src_top = highpart(arg1);
        src_bot = arg1;
    } else {
        dest_top = arg0;
        dest_bot = highpart(arg0);
        src_top = arg1;
        src_bot = highpart(arg1);
    }

    nonzero_label = gen_new_label();
    gen_op_jmp_nz32(arg2, nonzero_label);
    gen_op_mov32(dest_bot, src_bot);
    gen_op_mov32(dest_top, src_top);
    end_label = gen_new_label();
    gen_op_jmp(end_label);
    gen_set_label(nonzero_label);
    tmp = gen_new_qreg(QMODE_I32);
    gen_op_sub32(tmp, arg2, gen_im32(33));
    small_label = gen_new_label();
    gen_op_jmp_s32(tmp, small_label);
    /* Shift >= 32 bits.  */
    gen_split_shift32(dest_bot, src_top, tmp, right, arithmetic);
    if (arithmetic)
        gen_op_sar32(dest_top, dest_bot, gen_im32(31));
    else
        gen_op_mov32(dest_top, gen_im32(0));
    gen_op_jmp(end_label);
    gen_set_label(small_label);
    /* Shift < 32 bits.  */
    gen_split_shift32(dest_bot, src_bot, arg2, right, arithmetic);
    tmp = gen_new_qreg(QMODE_I32);
    gen_op_sub32(tmp, gen_im32(32), arg2);
    carry = gen_new_qreg(QMODE_I32);
    gen_split_shift32(carry, src_top, tmp, !right, 0);
    gen_op_or32(dest_bot, dest_bot, carry);
    gen_split_shift32(dest_top, src_top, arg2, right, arithmetic);
    gen_set_label(end_label);
}

static inline void gen_op_shr64(long arg0, long arg1, long arg2)
{
    split_shift64(arg0, arg1, arg2, 1, 0);
}

static inline void gen_op_sar64(long arg0, long arg1, long arg2)
{
    split_shift64(arg0, arg1, arg2, 1, 1);
}

static inline void gen_op_shl64(long arg0, long arg1, long arg2)
{
    split_shift64(arg0, arg1, arg2, 0, 0);
}

static inline void gen_op_ror64(long arg0, long arg1, long arg2)
{
    cpu_abort(NULL, "splitting ror64 not implemented");
}

static inline void gen_op_rol64(long arg0, long arg1, long arg2)
{
    cpu_abort(NULL, "splitting rol64 not implemented");
}

static inline void gen_op_mul64(long arg0, long arg1, long arg2)
{
    cpu_abort(NULL, "splitting mul64 not implemented");
}

static inline void gen_op_ext32u64(long arg0, long arg1)
{
    if (arg0 != arg1)
        gen_op_mov32(arg0, arg1);
    gen_op_mov32(highpart(arg0), gen_im32(0));
}

static inline void gen_op_ext32s64(long arg0, long arg1)
{
    if (arg0 != arg1)
        gen_op_mov32(arg0, arg1);
    gen_op_sar32(highpart(arg0), arg0, gen_im32(31));
}

static inline void gen_op_mov64i(long arg0, long arg1, long arg2)
{
    gen_op_mov32(arg0, gen_im32(arg1));
    gen_op_mov32(highpart(arg0), gen_im32(arg2));
}
#endif /* HOST_LONG_BITS */

static inline int gen_im64(uint64_t val)
{
    int reg;
    reg = gen_new_qreg(QMODE_I64);
    gen_op_mov64i(reg, (uint32_t)val, val >> 32);
    return reg;
}

#ifndef HAVE_op_bic32
static inline void gen_op_bic32(long arg0, long arg1, long arg2)
{
    int tmp;
    tmp = gen_new_qreg(QMODE_I32);
    gen_op_not32(tmp, arg2);
    gen_op_and32(arg0, arg1, tmp);
}
#endif

#ifndef HAVE_op_bic64
static inline void gen_op_bic64(long arg0, long arg1, long arg2)
{
    int tmp;
    tmp = gen_new_qreg(QMODE_I64);
    gen_op_not64(tmp, arg2);
    gen_op_and64(arg0, arg1, tmp);
}
#endif

#include "softfloat-gen.h"
