/*
 *  PowerPC Host code generation
 * 
 *  Copyright (c) 2005 CodeSourcery, LLC
 *  Written by Paul Brook
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define HOST_REG_ENV HOST_REG_AREG0
#define HOST_REG_SP  HOST_REG_R1
static inline int get_arg_regno(int qreg)
{
    qemu_assert(qreg >= FIRST_HARD_QREG, "bad hard regno");
    return qreg - FIRST_HARD_QREG;
}

static inline long get_arg_im(int qreg)
{
    qemu_assert(gen_reginfo[qreg].type == QREGT_IM, "bad immediate qreg");
    return gen_reginfo[qreg].val;
}

#define arg_regno(n) get_arg_regno(ARG(n))
#define arg_im(n) get_arg_im(ARG(n))

#define ENCODE_D(op1, rd, ra, d) \
    OUT32(((op1) << 26) | ((rd) << 21) | ((ra) << 16) | ((d) & 0xffff))
#define ENCODE_X(op1, rd, ra, rb, op2) \
    OUT32(((op1) << 26) | ((rd) << 21) | ((ra) << 16) | ((rb) << 11) \
          | ((op2) << 1))
#define ENCODE_XFX(op1, rs, spr, op2) \
    ENCODE_X(qop1, rs, spr & 0x1f, spr >> 5, op2);
#define ENCODE_I(op, li, aa, lk) \
    OUT32(((op) << 26) | (((li) << 2) & 0x07fffffc) | ((aa) << 1) | (lk))
#define ENCODE_B(op, bo, bi, bd, aa, lk) \
    OUT32(((op) << 26) | ((bo) << 21) | ((bi) << 16) \
          | (((bd) << 2) & 0xfffc) | ((aa) << 1) | (lk))
#define ENCODE_M(op, rs, ra, rb, mb, me) \
    OUT32(((op) << 26) | ((rs) << 21) | ((ra) << 16) | ((rb) << 11) \
          | ((mb) << 6) | ((me) << 1))
  
QOP_CODE_START
QOP_START(mov32)
    int dest = ARG(0);
    int src = ARG(1);
    int reg;
    int val;

    if (src < FIRST_HARD_QREG) {
        val = gen_reginfo[src].val;
        reg = arg_regno(0);
        switch (gen_reginfo[src].type) {
        case QREGT_ENV:
            ENCODE_D(32, reg, HOST_REG_ENV, val); /* lwz */
            break;
        case QREGT_STACK:
            ENCODE_D(32, reg, HOST_REG_SP, val + 16); /* lwz */
            break;
        case QREGT_IM:
            if ((int32_t)val != (int16_t)val) {
                /* lis reg, im */
                ENCODE_D(15, reg, HOST_REG_R0, val >> 16);
                /* ori reg, reg, im */
                ENCODE_D(24, reg, reg, val);
            } else {
                /* li reg, im */
                ENCODE_D(14, reg, HOST_REG_R0, val);
            }
            break;
        default:
            cpu_abort (NULL, "bad addressing mode");
        }
    } else if (dest < FIRST_HARD_QREG) {
        val = gen_reginfo[dest].val;
        reg = arg_regno(1);
        switch (gen_reginfo[dest].type) {
        case QREGT_ENV:
            ENCODE_D(36, reg, HOST_REG_ENV, val); /* stw */
            break;
        case QREGT_STACK:
            ENCODE_D(36, reg, HOST_REG_SP, val + 16); /* stw */
            break;
        default:
            cpu_abort (NULL, "bad addressing mode");
        }
    } else {
        reg = arg_regno(0);
        val = arg_regno(1);
        ENCODE_X(0x1f, val, reg, val, 0x1bc); /* mr */
    }
QOP_END

QOP_START(neg32)
    ENCODE_X(31, arg_regno(0), arg_regno(1), 0, 104);
QOP_END

QOP_START(not32)
    int src = arg_regno(1);
    /* nor dest, src, src */
    ENCODE_X(31, src, arg_regno(0), src, 124);
QOP_END

QOP_START(bswap32)
  int dest = arg_regno(0);
  int src = arg_regno(1);
  int tmp = arg_regno(2);
  /* rlwinm dest,src,24,16,23 */
  ENCODE_M(21, src, dest, 24, 16, 23);
  /* srwi tmp,src,24 */
  ENCODE_M(21, src, tmp, 8, 24, 31);
  /* or dest,dest,tmp */
  ENCODE_X(31, dest, dest, tmp, 444);
  /* rlwinm tmp,src,8,8,15 */
  ENCODE_M(21, src, tmp, 8, 8, 15);
  /* or dest,dest,tmp */
  ENCODE_X(31, dest, dest, tmp, 444);
  /* slwi tmp,src,24 */
  ENCODE_M(21, src, tmp, 24, 8, 15);
  /* or dest,dest,tmp */
  ENCODE_X(31, dest, dest, tmp, 444);
QOP_END

QOP_START(and32)
  ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 28);
QOP_END

QOP_START(or32)
  ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 444);
QOP_END

QOP_START(xor32)
  ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 316);
QOP_END

QOP_START(add32)
  ENCODE_X(31, arg_regno(0), arg_regno(2), arg_regno(1), 266);
QOP_END

QOP_START(sub32)
  ENCODE_X(31, arg_regno(0), arg_regno(2), arg_regno(1), 40);
QOP_END

QOP_START(shr32)
  int count = ARG(2);
  if (count < FIRST_HARD_QREG) {
      int im = arg_im(2);
      ENCODE_M(21, arg_regno(1), arg_regno(0), (32 - im) & 0x1f, im, 31);
  } else {
      ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 536);
  } 
QOP_END

QOP_START(sar32)
  int count = ARG(2);
  if (count < FIRST_HARD_QREG) {
      int im = arg_im(2);
      ENCODE_X(31, arg_regno(1), arg_regno(0), im, 824);
  } else {
      ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 792);
  } 
QOP_END

QOP_START(shl32)
  int count = ARG(2);
  if (count < FIRST_HARD_QREG) {
      int im = arg_im(2);
      ENCODE_M(21, arg_regno(1), arg_regno(0), im, 0, 31 - im);
  } else {
      ENCODE_X(31, arg_regno(1), arg_regno(0), arg_regno(2), 24);
  } 
QOP_END

QOP_START(rol32)
  int count = ARG(2);
  if (count < FIRST_HARD_QREG) {
      int im = arg_im(2);
      ENCODE_M(21, arg_regno(1), arg_regno(0), im, 0, 31);
  } else {
      ENCODE_X(23, arg_regno(1), arg_regno(0), arg_regno(2), 31);
  } 
QOP_END

/* ??? Implement this in gneric code.  */
QOP_START(ror32)
  int count = ARG(2);
  if (count < FIRST_HARD_QREG) {
      int im = arg_im(2);
      ENCODE_M(21, arg_regno(1), arg_regno(0), (32 - im) & 0x1f, 0, 31);
  } else {
      int tmp = arg_regno(3);
      /* subfic tmp, count, 32 */
      ENCODE_D(8, tmp, arg_regno(2), 32);
      /* rol dest, src, tmp */
      ENCODE_X(23, arg_regno(1), arg_regno(0), tmp, 31);
  } 
QOP_END

QOP_START(mul32)
  ENCODE_X(31, arg_regno(0), arg_regno(1), arg_regno(2), 235);
QOP_END

QOP_START(ext8s32)
  ENCODE_X(31, arg_regno(1), arg_regno(0), 0, 954);
QOP_END

QOP_START(ext16s32)
  ENCODE_X(31, arg_regno(1), arg_regno(0), 0, 922);
QOP_END

QOP_START(ext8u32)
  /* rlwinm dest, src, 0, 24, 31 */
  ENCODE_M(21, arg_regno(1), arg_regno(0), 0, 24, 31);
QOP_END

QOP_START(ext16u32)
  /* rlwinm dest, src, 0, 16, 31 */
  ENCODE_M(21, arg_regno(1), arg_regno(0), 0, 16, 31);
QOP_END

#define DO_SETcc32(name, negate, cond) \
QOP_START(set_##name##32) \
    int dest = arg_regno(0); \
    /* li dest, 1 */ \
    ENCODE_D(14, dest, 0, 1); \
    /* cmpw cr0, src1, src2 */ \
    ENCODE_X(31, 0, arg_regno(1), arg_regno(2), 0); \
    /* b<cc> dest */ \
    ENCODE_B(16, negate ? 4 : 12, cond, 2, 0, 0); \
    /* li dest, 0 */ \
    ENCODE_D(14, dest, 0, 0); \
QOP_END

DO_SETcc32(lt, 0, 0);
DO_SETcc32(le, 1, 1);

#define DO_JMPcc32(name, negate, cond) \
QOP_START(jmp_##name##32) \
    /* cmpwi cr0, src, 0 */ \
    ENCODE_D(11, 0, arg_regno(0), 0); \
    /* b<cc> dest */ \
    ENCODE_B(16, negate ? 4 : 12, cond, LABEL_OFFSET(ARG(1)) >> 2, 0, 0); \
QOP_END

DO_JMPcc32(z, 0, 2)
DO_JMPcc32(nz, 1, 2)
DO_JMPcc32(s, 0, 0)
DO_JMPcc32(ns, 1, 0)

QOP_START(jmp)
  ENCODE_I(18, LABEL_OFFSET(ARG(0)) >> 2, 0, 0);
QOP_END

QOP_START(start_tb)
    uint32_t val = ARG(0);
    
    /* ??? We probably also want to create a stack frame if we call functions
       directly.  */
    if (val) {
        /* Add two words for the frame header, and round up to a 16-byte
           boundary.  */
        val = (val + 8 + 15) & ~15;
        /* Create a stack frame.  */
        /* mflr r0 */
        OUT32(0x7c080286);
        /* stuw r1, -val(r1) */
        ENCODE_D(37, HOST_REG_SP, HOST_REG_SP, -val);
        /* stw r0, -(val + 4)(r1) */
        ENCODE_D(36, HOST_REG_R0, HOST_REG_SP, -(val + 4));
    }
QOP_END

QOP_START(goto_tb)
    uint32_t  stack_adj;
    stack_adj = ARG(0);
    /* ??? Need to make this restore lr if we do any direct calls.  */
#ifdef USE_DIRECT_JUMP
    MARK_JMP(ARG(1));
    if (stack_adj) {
        /* b end */
        ENCODE_I(18, 3, 0, 0);
        /* Free the stack frame.  */
        stack_adj = (stack_adj + 8 + 15) & ~15;
        /* lwz r0, (stack_adj + 4)(r1) */
        ENCODE_D(14, HOST_REG_SP, HOST_REG_R0, stack_adj);
        /* b <dest> */
        OUT32(0);
    } else {
        OUT32(0x60000000); /* nop */
    }
#else
    long addr;
    addr = ARG(2) + offsetof(TranslationBlock, tb_next[ARG(1)]);
    /* load_im reg, addr */
    ENCODE_LDIM(SCRATCH, addr)
    /* lwz reg, (reg) */
    ENCODE_D(32, SCRATCH, SCRATCH, 0);
    /* mtctr reg */
    ENCODE_XFX(31, SCRATCH, 9, 467);
    /* cmpi cr0, reg, 0 */
    ENCODE_D(11, 0, SCRATCH, 0);
    if (stack_adj) {
        /* be cr0,end */
        ENCODE_B(16, 12, 2, 3, 0, 0);
        /* Free the stack frame.  */
        stack_adj = (stack_adj + 8 + 15) & ~15;
        /* addi r1, stack_adj */
        ENCODE_D(14, HOST_REG_SP, HOST_REG_R0, stack_adj);
        /* bctr */
        ENCODE_X(19, 0x14, 0, 0, 528);
    } else {
        /* bnectr cr0 */
        ENCODE_X(19, 5, 2, 0, 528);
    }
#endif
QOP_END

QOP_START(exit_tb)
    uint32_t val = ARG(0);
    
    if (val) {
        /* Free the stack frame.  */
        val = (val + 8 + 15) & ~15;
        /* addi r1, val */
        ENCODE_D(14, HOST_REG_SP, HOST_REG_R0, val & 0xffff);
    }
    /* blr */
    OUT32(0x4e800020);
QOP_END

QOP_START(add64_split)
    /* addc dest_low, src1_low, src2_low */
    ENCODE_X(31, arg_regno(0), arg_regno(2), arg_regno(4), 10);
    /* adde dest_high, src1_high, src2_high */
    ENCODE_X(31, arg_regno(1), arg_regno(3), arg_regno(5), 138);
QOP_END

QOP_START(sub64_split)
    /* subfc dest_low, src2_low, src1_low */
    ENCODE_X(31, arg_regno(0), arg_regno(4), arg_regno(2), 8);
    /* subfe dest_high, src2_high, src1_high */
    ENCODE_X(31, arg_regno(1), arg_regno(5), arg_regno(3), 136);
QOP_END

QOP_START(neg64_split)
    /* subfic dest_low, src_low, 0 */
    ENCODE_D(8, arg_regno(0), arg_regno(2), 0);
    /* subfze dest_high, src_high */
    ENCODE_X(31, arg_regno(1), arg_regno(3), 0, 200);
QOP_END
QOP_CODE_END
