#! /bin/sh
# Script to generate list of qemu helper functions

set -e

if test -z "$1"; then
  echo "Usage: qop-helper.sh <srcfiles>"
  exit 1
fi
cat > qop-helper.h << EOF
/* Automatically generated by qop-helper.sh - do not modify */
#ifndef QOP_HELPER_H
#define DECLARE_HELPER32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint32_t arg0)
#define DECLARE_HELPER32_32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint32_t arg0, uint32_t arg1)
/* Floating point helpers.  */
/* ??? Fix these for HOST_FLOAT32 and HOST_FLOAT64.  */
#if HOST_LONG_BITS < 64
#define DECLARE_HELPERF64(name) void HELPER_PROTO helper_##name(struct CPUState *env, uint32_t retarg, uint32_t arg0)
#define DECLARE_HELPERF64x32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint32_t arg0)
#define DECLARE_HELPER32xF64(name) void HELPER_PROTO helper_##name(struct CPUState *env, uint32_t retarg, uint32_t arg0)
#define DECLARE_HELPERF64_F64(name) void HELPER_PROTO helper_##name(struct CPUState *env, uint32_t retarg, uint32_t arg0, uint32_t arg1)
#define DECLARE_HELPERF64_F64x32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint32_t arg0, uint32_t arg1)
#else
#define DECLARE_HELPERF64(name) uint64_t HELPER_PROTO helper_##name(struct CPUState *env, uint64_t retarg, uint64_t arg0)
#define DECLARE_HELPERF64x32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint64_t arg0)
#define DECLARE_HELPER32xF64(name) uint64_t HELPER_PROTO helper_##name(struct CPUState *env, uint32_t arg0)
#define DECLARE_HELPERF64_F64(name) uint64_t HELPER_PROTO helper_##name(struct CPUState *env, uint64_t arg0, uint64_t arg1)
#define DECLARE_HELPERF64_F64x32(name) uint32_t HELPER_PROTO helper_##name(struct CPUState *env, uint64_t arg0, uint64_t arg1)
#endif
enum qemu_helpers {
#define DEF(x) HELPER_##x,
#include "qop-helper.def"
#undef DEF
NUM_QOP_HELPERS
};
EOF
echo "/* Automatically generated by qop-helper.sh - do not modify */" \
  > qop-helper.def
sed -e '/^DECLARE_HELPER/!d;s/^DECLARE_HELPER\(.*\)(\(.*\))/\1 \2/' "$@" | \
while read type name; do
  echo "DEF($name)" >> qop-helper.def
  echo "DECLARE_HELPER$type($name);" >> qop-helper.h
done
echo "#endif" >> qop-helper.h
