enum {
    FIRST_TARGET_OP = NB_OPS,
#define DEF(name, nargs, barrier) INDEX_op_##name,
#include "qop-target.def"
#undef DEF
    LAST_TARGET_OP
};

static inline void gen_op0(int opcode, int barrier)
{
    qOP *new_qop;
    new_qop = allocate_qop(0, opcode);
    if (!gen_last_qop) {
        gen_first_qop = new_qop;
    } else {
        new_qop->prev = gen_last_qop;
        gen_last_qop->next = new_qop;
    }
    gen_last_qop = new_qop;
    if (barrier && !gen_throws_exception)
        gen_throws_exception = new_qop;
}

static inline void gen_op1(int opcode, int barrier, uint32_t arg0)
{
    qOP *new_qop;
    new_qop = allocate_qop(1, opcode);
    new_qop->args[0] = arg0;
    if (!gen_last_qop) {
        gen_first_qop = new_qop;
    } else {
        new_qop->prev = gen_last_qop;
        gen_last_qop->next = new_qop;
    }
    gen_last_qop = new_qop;
    if (barrier && !gen_throws_exception)
        gen_throws_exception = new_qop;
}

static inline void gen_op2(int opcode, int barrier, uint32_t arg0,
                           uint32_t arg1)
{
    qOP *new_qop;
    new_qop = allocate_qop(2, opcode);
    new_qop->args[0] = arg0;
    new_qop->args[1] = arg1;
    if (!gen_last_qop) {
        gen_first_qop = new_qop;
    } else {
        new_qop->prev = gen_last_qop;
        gen_last_qop->next = new_qop;
    }
    gen_last_qop = new_qop;
    if (barrier && !gen_throws_exception)
        gen_throws_exception = new_qop;
}

static inline void gen_op3(int opcode, int barrier, uint32_t arg0,
                           uint32_t arg1, uint32_t arg2)
{
    qOP *new_qop;
    new_qop = allocate_qop(3, opcode);
    new_qop->args[0] = arg0;
    new_qop->args[1] = arg1;
    new_qop->args[2] = arg2;
    if (!gen_last_qop) {
        gen_first_qop = new_qop;
    } else {
        new_qop->prev = gen_last_qop;
        gen_last_qop->next = new_qop;
    }
    gen_last_qop = new_qop;
    if (barrier && !gen_throws_exception)
        gen_throws_exception = new_qop;
}

#define DEF0(name, barrier) \
static inline void gen_op_##name(void) \
{ gen_op0(INDEX_op_##name, barrier); }
#define DEF1(name, barrier) \
static inline void gen_op_##name(uint32_t arg0) \
{ gen_op1(INDEX_op_##name, barrier, arg0); }
#define DEF2(name, barrier) \
static inline void gen_op_##name(uint32_t arg0, uint32_t arg1) \
{ gen_op2(INDEX_op_##name, barrier, arg0, arg1); }
#define DEF3(name, barrier) \
static inline void gen_op_##name(uint32_t arg0, uint32_t arg1, uint32_t arg2) \
{ gen_op3(INDEX_op_##name, barrier, arg0, arg1, arg2); }
#define DEF(name, nargs, barrier) DEF##nargs(name, barrier)
#include "qop-target.def"
#undef DEF
#undef DEF0
#undef DEF1
#undef DEF3

static inline void remove_qop(qOP *qop)
{
  if (qop->next)
      qop->next->prev = qop->prev;
  else
      gen_last_qop = qop->prev;
  if (qop->prev)
      qop->prev->next = qop->next;
  else
      gen_first_qop = qop->next;
}
