
enum target_qregs {
    QREG_NULL,
#define DEFO32(name, offset) QREG_ ## name,
#define DEFO64(name, offset) DEFO32(name, offset)
#define DEFF32(name, reg) DEFO32(name, reg)
#define DEFF64(name, reg) DEFO32(name, reg)
#define DEFR(name, reg, mode) DEFO32(name, reg)
#include "qregs.def"
#undef DEFO64
#undef DEFO32
#undef DEFF64
#undef DEFF32
#undef DEFR
#if HOST_LONG_BITS < 64
#define DEFO32(name, offset)
#define DEFO64(name, offset) QREGHI_ ## name,
#define DEFF32(name, reg)
#define DEFF64(name, reg)
#define DEFR(name, reg, mode)
#include "qregs.def"
#undef DEFO64
#undef DEFO32
#undef DEFF64
#undef DEFF32
#undef DEFR
#endif
    TARGET_NUM_REGS
};

/* ??? No idea what a sensible value for this is.  */
#define MAX_QREGS 2048
/* ??? It's hard to estimate how large the code for each qop will be.
   Empirically this seems to be a safe value.  */
#define MAX_QOPS 1024

enum {
#define DEF(s, n, copy_size) INDEX_op_ ## s,
#include "opc.h"
#undef DEF
    NB_OPS,
};

qOP *allocate_qop(int nargs, uint16_t opcode);

struct qreg gen_reginfo[MAX_QREGS];
int gen_last_qreg;

static inline int gen_new_qreg(qreg_mode mode)
{
    struct qreg *info;
    int n;

#if !defined(HOST_FLOAT32)
    if (mode == QMODE_F32)
        mode = QMODE_I32;
#endif
#if HOST_LONG_BITS < 64
    if (mode == QMODE_I64) {
        /* On 32-bit hosts 64-bt quantities are split into a pair of 32-bit
           values.  */
        if (gen_last_qreg + 1 >= MAX_QREGS)
            cpu_abort(NULL, "Too many qregs");
        info = &gen_reginfo[gen_last_qreg];
        memset(info, 0, 2 * sizeof(struct qreg));
        n = gen_last_qreg;
        gen_last_qreg += 2;
        info[0].mode = QMODE_I32;
        info[0].highpart = n + 1;
        info[1].mode = QMODE_I32;
        return n;
    }
#else
#if !defined(HOST_FLOAT64)
    if (mode == QMODE_F64)
        mode = QMODE_I64;
#endif
#endif
    if (gen_last_qreg >= MAX_QREGS)
        cpu_abort(NULL, "Too many qregs");
    info = &gen_reginfo[gen_last_qreg];
    memset(info, 0, sizeof(struct qreg));
    n = gen_last_qreg++;
    info->mode = mode;
    return n;
}

static inline int gen_im32(uint32_t val)
{
    int reg;
    reg = gen_new_qreg(QMODE_I32);
    gen_reginfo[reg].type = QREGT_IM;
    gen_reginfo[reg].val = val;
    return reg;
}

static inline void gen_insert_unwind(qOP *qop, uint32_t arg0, uint32_t arg1,
                                     uint32_t arg2)
{
    qOP *new_qop;

    new_qop = allocate_qop(3, INDEX_op_unwind);
    new_qop->args[0] = arg0;
    new_qop->args[1] = arg1;
    new_qop->args[2] = arg2;
    new_qop->next = qop;
    new_qop->prev = qop->prev;
    if (qop->prev) {
        qop->prev->next = new_qop;
    } else {
        gen_first_qop = new_qop;
    }
    qop->prev = new_qop;
}

/* ??? Put this somethere else.  */
#define qemu_assert(cond, msg) \
  do { if (!(cond)) cpu_abort(NULL, msg); } while(0)
