/* Wrappers for soft-float helpers.  */
#include "qop-helper.h"

#if HOST_LONG_BITS < 64
#define gen_float64_unary_helper gen_op_helperf64
#define gen_float64_binary_helper gen_op_helperf64_f64
#define gen_float64_compare_helper gen_op_helperf64_f64x32
#define gen_float64_ret32 gen_op_helperf64x32
#define gen_float64_arg32 gen_op_helper32xf64
#else
#define gen_float64_unary_helper gen_op_helper64
#define gen_float64_binary_helper gen_op_helper64_64
#define gen_float64_compare_helper gen_op_helper64_64x32
#define gen_float64_ret32 gen_op_helper64x32
#define gen_float64_arg32 gen_op_helper32x64
#endif

#define gen_binary32(op) \
static inline void gen_op_##op##f32(int arg0, int arg1, int arg2) \
{ \
    gen_op_helper32_32(arg0, arg1, arg2, HELPER_float32_##op); \
}

#define gen_unary32(op) \
static inline void gen_op_##op##f32(int arg0, int arg1) \
{ \
    gen_op_helper32(arg0, arg1, HELPER_float32##_##op); \
}

#define gen_binary64(op) \
static inline void gen_op_##op##f64(int arg0, int arg1, int arg2) \
{ \
    gen_float64_binary_helper(arg0, arg1, arg2, HELPER_float64_##op); \
}

#define gen_unary64(op) \
static inline void gen_op_##op##f64(int arg0, int arg1) \
{ \
    gen_float64_unary_helper(arg0, arg1, HELPER_float64##_##op); \
}

#if !defined(HOST_FLOAT32)
gen_binary32(add)
gen_binary32(sub)
gen_binary32(mul)
gen_binary32(div)

gen_unary32(abs)
gen_unary32(sqrt)
gen_unary32(chs)
gen_unary32(round_to_int)

static inline void gen_op_movf32(int arg0, int arg1)
{
    return gen_op_mov32(arg0, arg1);
}

static inline void gen_op_ldf32(int arg0, int arg1)
{
    gen_op_ld32(arg0, arg1);
}

static inline void gen_op_stf32(int arg0, int arg1)
{
    gen_op_st32(arg0, arg1);
}

static inline void gen_op_iround_f32(int arg0, int arg1)
{
    gen_op_helper32(arg0, arg1, HELPER_float32_round_to_int);
}

static inline void gen_op_itrunc_f32(int arg0, int arg1)
{
    gen_op_helper32(arg0, arg1, HELPER_float32_trunc_to_int);
}

static inline void gen_op_i32_to_f32(int arg0, int arg1)
{
    gen_op_helper32(arg0, arg1, HELPER_int32_to_float32);
}

static inline void gen_op_f32_to_i32(int arg0, int arg1)
{
    gen_op_helper32(arg0, arg1, HELPER_float32_to_int32);
}

static inline void gen_op_f32_to_f64(int arg0, int arg1)
{
    gen_float64_arg32(arg0, arg1, HELPER_float32_to_float64);
}

static inline void gen_op_zerof32(int arg0)
{
    gen_op_mov32(arg0, gen_im32(0));
}

static inline void gen_op_pack_f32_32(int arg0, int arg1)
{
    gen_op_mov32(arg0, arg1);
}

static inline void gen_op_pack_32_f32(int arg0, int arg1)
{
    gen_op_mov32(arg0, arg1);
}

#endif /* !HOST_FLOAT32 */

#if !defined(HOST_FLOAT64)

gen_binary32(compare)
gen_binary32(compare_quiet)

gen_unary64(abs)
gen_unary64(sqrt)
gen_unary64(chs)

gen_binary64(add)
gen_binary64(sub)
gen_binary64(mul)
gen_binary64(div)

static inline void gen_op_iround_f64(int arg0, int arg1)
{
    gen_float64_unary_helper(arg0, arg1, HELPER_float64_round_to_int);
}

static inline void gen_op_itrunc_f64(int arg0, int arg1)
{
    gen_float64_unary_helper(arg0, arg1, HELPER_float64_trunc_to_int);
}

static inline void gen_op_comparef64(int arg0, int arg1, int arg2)
{
    gen_float64_compare_helper(arg0, arg1, arg2, HELPER_float64_compare);
}

static inline void gen_op_compare_quietf64(int arg0, int arg1, int arg2)
{
    gen_float64_compare_helper(arg0, arg1, arg2, HELPER_float64_compare_quiet);
}

static inline void gen_op_i32_to_f64(int arg0, int arg1)
{
    gen_float64_arg32(arg0, arg1, HELPER_int32_to_float64);
}

static inline void gen_op_f64_to_i32(int arg0, int arg1)
{
    gen_float64_ret32(arg0, arg1, HELPER_float64_to_int32);
}

static inline void gen_op_f64_to_f32(int arg0, int arg1)
{
    gen_float64_ret32(arg0, arg1, HELPER_float64_to_float32);
}

#ifndef HAVE_op_zerof64
static inline void gen_op_zerof64(int arg0)
{
    gen_float64_arg32(arg0, gen_im32(0), HELPER_int32_to_float64);
}
#endif

#if HOST_LONG_BITS == 64
static inline void gen_op_ldf64(int arg0, int arg1)
{
    gen_op_ld64(arg0, arg1);
}
static inline void gen_op_stf64(int arg0, int arg1)
{
    gen_op_st64(arg0, arg1);
}
static inline void gen_op_movf64(int arg0, int arg1)
{
    gen_op_mov64(arg0, arg1);
}
#define gen_op_helperf64_f64 gen_op_helper64_64
#endif
#endif /* !HOST_FLOAT64 */

#undef gen_unary32
#undef gen_unary64
#undef gen_binary32
#undef gen_binary64
