/* Soft-float op helper routines.  */
/* 32-bit floats are treated like 32-bit integers, with helper routines to
   do the actual arithmetic.
   On 64-bit hosts 64-bit floats are treated similarly.
   On 32-bit hosts we have an array of pseudo registers in memory, and the
   host register numbers are indices into this array.  Helper functions
   are passed the register number, not the value.  */
/* ??? This needs splitting up to handle HOST_FLOAT32 and HOST_FLOAT64
 * separately if/when targets that need this exist.  */
#include <stdio.h>

#include "config.h"
#include "cpu.h"
#include "exec-all.h"
#include "qop-helper.h"
#include "softfloat.h"
#include "softfloat-helper.h"

#define binop32(op) \
    returnf32(float32_##op(argf32(arg0), argf32(arg1), &CPU_FP_STATUS))

/* 32-bit operations.  */
DECLARE_HELPER32_32(float32_add)
{
    binop32(add);
}

DECLARE_HELPER32_32(float32_sub)
{
    binop32(sub);
}

DECLARE_HELPER32_32(float32_mul)
{
    binop32(mul);
}

DECLARE_HELPER32_32(float32_div)
{
    binop32(div);
}

DECLARE_HELPER32_32(float32_compare)
{
    binop32(compare);
}

DECLARE_HELPER32_32(float32_compare_quiet)
{
    binop32(compare_quiet);
}

DECLARE_HELPER32(float32_abs)
{
    returnf32(float32_abs(argf32(arg0)));
}

DECLARE_HELPER32(float32_chs)
{
    returnf32(float32_chs(argf32(arg0)));
}

DECLARE_HELPER32(float32_sqrt)
{
    returnf32(float32_sqrt(argf32(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPER32(float32_round_to_int)
{
    returnf32(float32_round_to_int(argf32(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPER32(float32_trunc_to_int)
{
    returnf32(float32_trunc_to_int(argf32(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPER32(int32_to_float32)
{
    returnf32(int32_to_float32((int32_t)arg0, &CPU_FP_STATUS));
}

DECLARE_HELPER32(float32_to_int32)
{
    return float32_to_int32(argf32(arg0), &CPU_FP_STATUS);
}
    
DECLARE_HELPER32xF64(float32_to_float64)
{
    returnf64(float32_to_float64(argf32(arg0), &CPU_FP_STATUS));
}

/* 64-bit operations.  */
#define binop64(op) \
    returnf64(float64_##op(argf64(arg0), argf64(arg1), &CPU_FP_STATUS))

DECLARE_HELPERF64_F64(float64_add)
{
    binop64(add);
}

DECLARE_HELPERF64_F64(float64_sub)
{
    binop64(sub);
}

DECLARE_HELPERF64_F64(float64_mul)
{
    binop64(mul);
}

DECLARE_HELPERF64_F64(float64_div)
{
    binop64(div);
}

DECLARE_HELPERF64_F64x32(float64_compare)
{
    return float64_compare(argf64(arg0), argf64(arg1), &CPU_FP_STATUS);
}

DECLARE_HELPERF64_F64x32(float64_compare_quiet)
{
    return float64_compare_quiet(argf64(arg0), argf64(arg1), &CPU_FP_STATUS);
}

DECLARE_HELPERF64(float64_abs)
{
    returnf64(float64_abs(argf64(arg0)));
}

DECLARE_HELPERF64(float64_chs)
{
    returnf64(float64_chs(argf64(arg0)));
}

DECLARE_HELPERF64(float64_sqrt)
{
    returnf64(float64_sqrt(argf64(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPERF64(float64_round_to_int)
{
    returnf64(float64_round_to_int(argf64(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPERF64(float64_trunc_to_int)
{
    returnf64(float64_trunc_to_int(argf64(arg0), &CPU_FP_STATUS));
}

DECLARE_HELPER32xF64(int32_to_float64)
{
    returnf64(int32_to_float64((int32_t)arg0, &CPU_FP_STATUS));
}

DECLARE_HELPERF64x32(float64_to_int32)
{
    return float64_to_int32(argf64(arg0), &CPU_FP_STATUS);
}

DECLARE_HELPERF64x32(float64_to_float32)
{
    returnf32(float64_to_float32(argf64(arg0), &CPU_FP_STATUS));
}

/* TODO: float80.  Can implement same as float64 on 32-bit hosts.  */
