/* Header file for floating point helper routines.  */
#ifndef SOFTFLOAT_HELPER_H
#define SOFTFLOAT_HELPER_H

#define CPU_FP_STATUS env->fp_status

#ifdef HOST_FLOAT32

#define argf32(x) (x)
#define returnf32(x) return (x)

#else
static inline uint32_t f2i32(float32 f)
{
    union {
        uint32_t i;
        float32 f;
    } v;
    v.f = f;
    return v.i;
}
#define returnf32(x) return f2i32(x)
static inline float32 argf32(uint32_t i)
{
    union {
        uint32_t i;
        float32 f;
    } v;
    v.i = i;
    return v.f;
}
#endif

#ifdef HOST_FLOAT64

#define argf64(x) (x)
#define returnf64(x) return (x)

#else
#if HOST_LONG_BITS < 64

#define argf64(x) env->float64regs[x]
#define returnf64(val) env->float64regs[retarg] = (val)

#else /* HOST_LONG_BITS >= 64 */

static inline float64 argf64(uint64_t i)
{
    union {
        uint64_t i;
        float64 f;
    } v;
    v.i = i;
    return v.f;
}

static inline uint64_t f2i64(float64 f)
{
    union {
        uint64_t i;
        float64 f;
    } v;
    v.f = f;
    return v.i;
}

#define returnf64(val) return f2i64(val);
#endif

#endif


#endif /* SOFTFLOAT_HELPER_H */


